@@ -1,11 +1,10 @@
-/*
- * -----------------------------------------------------------------
- * Programmer(s): Daniel Reynolds @ SMU
- *                Scott D. Cohen, Alan C. Hindmarsh, Radu Serban
- *                   and Dan Shumaker @ LLNL
+/* -----------------------------------------------------------------
+ * Programmer(s): Scott D. Cohen, Alan C. Hindmarsh, Radu Serban and
+ *                Dan Shumaker @ LLNL
+ *                Daniel Reynolds @ SMU
  * -----------------------------------------------------------------
  * SUNDIALS Copyright Start
- * Copyright (c) 2002-2020, Lawrence Livermore National Security
+ * Copyright (c) 2002-2022, Lawrence Livermore National Security
  * and Southern Methodist University.
  * All rights reserved.
  *
@@ -15,26 +14,45 @@
  * SUNDIALS Copyright End
  * -----------------------------------------------------------------
  * Implementation header file for the main CVODE integrator.
- * -----------------------------------------------------------------
- */
+ * -----------------------------------------------------------------*/
 
 #ifndef _CVODE_IMPL_H
 #define _CVODE_IMPL_H
 
 #include <stdarg.h>
+
 #include <cvode/cvode.h>
+#include "cvode_proj_impl.h"
+#include "sundials/sundials_context_impl.h"
+#include "sundials/sundials_logger_impl.h"
+#include "sundials/sundials_math.h"
 
 #ifdef __cplusplus  /* wrapper to enable C++ usage */
 extern "C" {
 #endif
 
+#if defined(SUNDIALS_EXTENDED_PRECISION)
+#define RSYM  ".32Lg"
+#define RSYMW "19.32Lg"
+#else
+#define RSYM  ".16g"
+#define RSYMW "23.16g"
+#endif
+
+/*=================================================================*/
+/* Shortcuts                                                       */
+/*=================================================================*/
+
+#define CV_PROFILER cv_mem->cv_sunctx->profiler
+#define CV_LOGGER cv_mem->cv_sunctx->logger
+
 /*
  * =================================================================
- *   M A I N    I N T E G R A T O R    M E M O R Y    B L O C K
+ *   I N T E R N A L   C O N S T A N T S
  * =================================================================
  */
 
-/* Basic CVODE constants */
+/* Basic constants */
 
 #define ADAMS_Q_MAX 12     /* max value of q for lmm == ADAMS     */
 #define BDF_Q_MAX    5     /* max value of q for lmm == BDF       */
@@ -47,14 +65,132 @@ extern "C" {
 #define MXHNIL_DEFAULT   10             /* mxhnil default value   */
 #define MXSTEP_DEFAULT   500            /* mxstep default value   */
 
-/* Return values for lower level routines used by CVode and functions
-   provided to the nonlinear solver */
+#define MSBP_DEFAULT         20          /* max steps between lsetup calls */
+#define DGMAX_LSETUP_DEFAULT RCONST(0.3) /* gamma threshold to call lsetup */
+
+/* Step size change constants
+ * --------------------------
+ * ETA_MIN_FX_DEFAULT  if eta_min_fx < eta < eta_max_fx reject a change in step size or order
+ * ETA_MAX_FX_DEFAULT
+ * ETA_MAX_FS_DEFAULT  -+
+ * ETA_MAX_ES_DEFAULT   |
+ * ETA_MAX_GS_DEFAULT   |
+ * ETA_MIN_DEFAULT      |-> bounds for eta (step size change)
+ * ETA_MAX_EF_DEFAULT   |
+ * ETA_MIN_EF_DEFAULT   |
+ * ETA_CF_DEFAULT      -+
+ * SMALL_NST_DEFAULT   nst <= SMALL_NST => use eta_max_es
+ * SMALL_NEF_DEFAULT   if small_nef <= nef <= MXNEF1, then eta =  SUNMIN(eta, eta_max_ef)
+ * ONEPSM (1+epsilon)  used in testing if the step size is below its bound
+ */
+
+#define ETA_MIN_FX_DEFAULT RCONST(0.0)
+#define ETA_MAX_FX_DEFAULT RCONST(1.5)
+#define ETA_MAX_FS_DEFAULT RCONST(10000.0)
+#define ETA_MAX_ES_DEFAULT RCONST(10.0)
+#define ETA_MAX_GS_DEFAULT RCONST(10.0)
+#define ETA_MIN_DEFAULT    RCONST(0.1)
+#define ETA_MAX_EF_DEFAULT RCONST(0.2)
+#define ETA_MIN_EF_DEFAULT RCONST(0.1)
+#define ETA_CF_DEFAULT     RCONST(0.25)
+#define SMALL_NST_DEFAULT  10
+#define SMALL_NEF_DEFAULT  2
+#define ONEPSM             RCONST(1.000001)
+
+/* Step size controller constants
+ * ------------------------------
+ * ADDON  safety factor in computing eta
+ * BIAS1  -+
+ * BIAS2   |-> bias factors in eta selection
+ * BIAS3  -+
+ */
+
+#define ADDON     RCONST(0.000001)
+#define BIAS1     RCONST(6.0)
+#define BIAS2     RCONST(6.0)
+#define BIAS3     RCONST(10.0)
+
+/* Order selection constants
+ * -------------------------
+ * LONG_WAIT   number of steps to wait before considering an order change when
+ *             q==1 and MXNEF1 error test failures have occurred
+ */
+
+#define LONG_WAIT    10
+
+/* Failure limits
+ * --------------
+ * MXNCF   max no. of convergence failures during one step try
+ * MXNEF   max no. of error test failures during one step try
+ * MXNEF1  max no. of error test failures before forcing a reduction of order
+ */
+
+#define MXNCF        10
+#define MXNEF         7
+#define MXNEF1        3
+
+/* Control constants for lower-level functions used by cvStep
+ * ----------------------------------------------------------
+ *
+ * cvHin return values:
+ *    CV_SUCCESS
+ *    CV_RHSFUNC_FAIL
+ *    CV_TOO_CLOSE
+ *
+ * cvStep control constants:
+ *    DO_ERROR_TEST
+ *    PREDICT_AGAIN
+ *
+ * cvStep return values:
+ *    CV_SUCCESS,
+ *    CV_LSETUP_FAIL,  CV_LSOLVE_FAIL,
+ *    CV_RHSFUNC_FAIL, CV_RTFUNC_FAIL
+ *    CV_CONV_FAILURE, CV_ERR_FAILURE,
+ *    CV_FIRST_RHSFUNC_ERR
+ *
+ * cvNls input nflag values:
+ *    FIRST_CALL
+ *    PREV_CONV_FAIL
+ *    PREV_PROJ_FAIL
+ *    PREV_ERR_FAIL
+ *
+ * cvNls return values:
+ *    CV_SUCCESS,
+ *    CV_LSETUP_FAIL, CV_LSOLVE_FAIL, CV_RHSFUNC_FAIL,
+ *    CONV_FAIL, RHSFUNC_RECVR
+ *
+ * cvNewtonIteration return values:
+ *    CV_SUCCESS,
+ *    CV_LSOLVE_FAIL, CV_RHSFUNC_FAIL
+ *    RHSFUNC_RECVR,
+ *    TRY_AGAIN
+ *
+ */
+
+#define DO_ERROR_TEST    +2
+#define PREDICT_AGAIN    +3
+
+#define TRY_AGAIN        +5
+#define FIRST_CALL       +6
+#define PREV_CONV_FAIL   +7
+#define PREV_PROJ_FAIL   +8
+#define PREV_ERR_FAIL    +9
+
+#define RHSFUNC_RECVR    +10
+#define CONSTR_RECVR     +11
+#define CONSTRFUNC_RECVR +12
+#define PROJFUNC_RECVR   +13
+
+/*
+ * =================================================================
+ *   M A I N    I N T E G R A T O R    M E M O R Y    B L O C K
+ * =================================================================
+ */
 
-#define RHSFUNC_RECVR +9
 
 /*
  * -----------------------------------------------------------------
- * Types : struct CVodeMemRec, CVodeMem
+ * Types: struct CVodeMemRec, CVodeMem
  * -----------------------------------------------------------------
  * The type CVodeMem is type pointer to struct CVodeMemRec.
  * This structure contains fields to keep track of problem state.
@@ -63,6 +199,8 @@ extern "C" {
 
 typedef struct CVodeMemRec {
 
+  SUNContext cv_sunctx;
+
   realtype cv_uround;    /* machine unit roundoff */
 
   /*--------------------------
@@ -91,20 +229,20 @@ typedef struct CVodeMemRec {
 
   N_Vector cv_zn[L_MAX];  /* Nordsieck array, of size N x (q+1).
                              zn[j] is a vector of length N (j=0,...,q)
-                             zn[j] = [1/factorial(j)] * h^j * (jth
-                             derivative of the interpolating polynomial       */
+                             zn[j] = [1/factorial(j)] * h^j *
+                             (jth derivative of the interpolating polynomial) */
 
-  /*--------------------------
-    other vectors of length N
-    -------------------------*/
+  /*-------------------
+    Vectors of length N
+    -------------------*/
 
   N_Vector cv_ewt;     /* error weight vector                                 */
   N_Vector cv_y;       /* y is used as temporary storage by the solver
                           The memory is provided by the user to CVode
                           where the vector is named yout.                     */
   N_Vector cv_acor;    /* In the context of the solution of the nonlinear
-                          equation, acor = y_n(m) - y_n(0). On return,
-                          this vector is scaled to give the est. local err.   */
+                          equation, acor = y_n(m) - y_n(0). On return, this
+                          vector is scaled to give the estimated local error  */
   N_Vector cv_tempv;   /* temporary storage vector                            */
   N_Vector cv_ftemp;   /* temporary storage vector                            */
   N_Vector cv_vtemp1;  /* temporary storage vector                            */
@@ -126,7 +264,7 @@ typedef struct CVodeMemRec {
 
   int cv_q;                    /* current order                               */
   int cv_qprime;               /* order to be used on the next step
-                                  = q-1, q, or q+1                            */
+                                  qprime = q-1, q, or q+1                     */
   int cv_next_q;               /* order to be used on the next step           */
   int cv_qwait;                /* number of internal steps to wait before
                                   considering a change in q                   */
@@ -139,7 +277,7 @@ typedef struct CVodeMemRec {
   realtype cv_eta;             /* eta = hprime / h                            */
   realtype cv_hscale;          /* value of h used in zn                       */
   realtype cv_tn;              /* current internal value of t                 */
-  realtype cv_tretlast;        /* value of tret last returned by CVode        */
+  realtype cv_tretlast;        /* last value of t returned by CVode           */
 
   realtype cv_tau[L_MAX+1];    /* array of previous q+1 successful step
                                   sizes indexed from 1 to q+1                 */
@@ -147,33 +285,43 @@ typedef struct CVodeMemRec {
                                   1 to NUM_TESTS(=5)                          */
   realtype cv_l[L_MAX];        /* coefficients of l(x) (degree q poly)        */
 
-  realtype cv_rl1;              /* the scalar 1/l[1]                          */
-  realtype cv_gamma;            /* gamma = h * rl1                            */
-  realtype cv_gammap;           /* gamma at the last setup call               */
-  realtype cv_gamrat;           /* gamma / gammap                             */
+  realtype cv_rl1;             /* the scalar 1/l[1]                           */
+  realtype cv_gamma;           /* gamma = h * rl1                             */
+  realtype cv_gammap;          /* gamma at the last setup call                */
+  realtype cv_gamrat;          /* gamma / gammap                              */
 
-  realtype cv_crate;            /* estimated corrector convergence rate       */
-  realtype cv_delp;             /* norm of previous nonlinear solver update   */
-  realtype cv_acnrm;            /* | acor | wrms                              */
-  booleantype cv_acnrmcur;      /* is | acor | wrms current?                  */
-  realtype cv_nlscoef;          /* coeficient in nonlinear convergence test   */
+  realtype cv_crate;           /* estimated corrector convergence rate        */
+  realtype cv_delp;            /* norm of previous nonlinear solver update    */
+  realtype cv_acnrm;           /* | acor |                                    */
+  booleantype cv_acnrmcur;     /* is | acor | current?                        */
+  realtype cv_nlscoef;         /* coeficient in nonlinear convergence test    */
 
   /*------
     Limits
     ------*/
 
-  int cv_qmax;          /* q <= qmax                                          */
-  long int cv_mxstep;   /* maximum number of internal steps for one user call */
-  int cv_maxcor;        /* maximum number of corrector iterations for the
-                           solution of the nonlinear equation                 */
-  int cv_mxhnil;        /* maximum number of warning messages issued to the
-                           user that t + h == t for the next internal step    */
-  int cv_maxnef;        /* maximum number of error test failures              */
-  int cv_maxncf;        /* maximum number of nonlinear convergence failures   */
+  int cv_qmax;            /* q <= qmax                                          */
+  long int cv_mxstep;     /* maximum number of internal steps for one user call */
+  int cv_mxhnil;          /* maximum number of warning messages issued to the
+                             user that t + h == t for the next internal step    */
+  int cv_maxnef;          /* maximum number of error test failures              */
+  int cv_maxncf;          /* maximum number of nonlinear convergence failures   */
 
-  realtype cv_hmin;     /* |h| >= hmin                                        */
-  realtype cv_hmax_inv; /* |h| <= 1/hmax_inv                                  */
-  realtype cv_etamax;   /* eta <= etamax                                      */
+  realtype cv_hmin;       /* |h| >= hmin                                        */
+  realtype cv_hmax_inv;   /* |h| <= 1/hmax_inv                                  */
+  realtype cv_etamax;     /* eta <= etamax                                      */
+  realtype cv_eta_min_fx; /* eta_min_fx < eta < eta_max_fx keep the current h   */
+  realtype cv_eta_max_fx;
+  realtype cv_eta_max_fs; /* eta <= eta_max_fs on the first step                */
+  realtype cv_eta_max_es; /* eta <= eta_max_es on early steps                   */
+  realtype cv_eta_max_gs; /* eta <= eta_max_gs on a general step                */
+  realtype cv_eta_min;    /* eta >= eta_min on a general step                   */
+  realtype cv_eta_min_ef; /* eta >= eta_min_ef after an error test failure      */
+  realtype cv_eta_max_ef; /* eta on multiple (>= small_nef) error test failures */
+  realtype cv_eta_cf;     /* eta on a nonlinear solver convergence failure      */
+
+  long int cv_small_nst; /* nst <= small_nst use eta_max_es */
+  int cv_small_nef;      /* nef >= small_nef use eta_max_ef */
 
   /*--------
     Counters
@@ -182,22 +330,27 @@ typedef struct CVodeMemRec {
   long int cv_nst;         /* number of internal steps taken                  */
   long int cv_nfe;         /* number of f calls                               */
   long int cv_ncfn;        /* number of corrector convergence failures        */
+  long int cv_nni;         /* number of nonlinear iterations performed        */
+  long int cv_nnf;         /* number of nonlinear convergence failures        */
   long int cv_netf;        /* number of error test failures                   */
-  long int cv_nni;         /* number of Newton iterations performed           */
   long int cv_nsetups;     /* number of setup calls                           */
   int cv_nhnil;            /* number of messages issued to the user that
                               t + h == t for the next iternal step            */
 
+  /*----------------
+    Step size ratios
+    ----------------*/
+
   realtype cv_etaqm1;      /* ratio of new to old h for order q-1             */
   realtype cv_etaq;        /* ratio of new to old h for order q               */
   realtype cv_etaqp1;      /* ratio of new to old h for order q+1             */
 
-  /*----------------------------
-    Space requirements for CVODE
-    ----------------------------*/
+  /*------------------
+    Space requirements
+    ------------------*/
 
-  sunindextype cv_lrw1;        /* no. of realtype words in 1 N_Vector             */
-  sunindextype cv_liw1;        /* no. of integer words in 1 N_Vector              */
+  sunindextype cv_lrw1;        /* no. of realtype words in 1 N_Vector y           */
+  sunindextype cv_liw1;        /* no. of integer words in 1 N_Vector y            */
   long int cv_lrw;             /* no. of realtype words in CVODE work vectors     */
   long int cv_liw;             /* no. of integer words in CVODE work vectors      */
 
@@ -205,11 +358,11 @@ typedef struct CVodeMemRec {
     Nonlinear Solver Data
     ---------------------*/
 
-  SUNNonlinearSolver NLS;  /* Sundials generic nonlinear solver object */
-  booleantype ownNLS;      /* flag indicating if CVODE created the nonlinear
-                              solver object */
-  int convfail;            /* flag to indicate when a Jacbian update may
-                              be needed */
+  SUNNonlinearSolver NLS;      /* nonlinear solver object                   */
+  booleantype ownNLS;          /* flag indicating NLS ownership             */
+  CVRhsFn nls_f;               /* f(t,y(t)) used in the nonlinear solver    */
+  int convfail;                /* flag to indicate when a Jacobian update may
+                                  be needed */
 
   /*------------------
     Linear Solver Data
@@ -219,18 +372,21 @@ typedef struct CVodeMemRec {
 
   int (*cv_linit)(struct CVodeMemRec *cv_mem);
 
-  int (*cv_lsetup)(struct CVodeMemRec *cv_mem, int convfail, N_Vector ypred,
-		   N_Vector fpred, booleantype *jcurPtr, N_Vector vtemp1,
-		   N_Vector vtemp2, N_Vector vtemp3);
+  int (*cv_lsetup)(struct CVodeMemRec *cv_mem, int convfail,
+                   N_Vector ypred, N_Vector fpred, booleantype *jcurPtr,
+                   N_Vector vtemp1, N_Vector vtemp2, N_Vector vtemp3);
 
   int (*cv_lsolve)(struct CVodeMemRec *cv_mem, N_Vector b, N_Vector weight,
-		   N_Vector ycur, N_Vector fcur);
+                   N_Vector ycur, N_Vector fcur);
 
   int (*cv_lfree)(struct CVodeMemRec *cv_mem);
 
   /* Linear Solver specific memory */
 
-  void *cv_lmem;
+  void     *cv_lmem;         /* linear solver interface memory structure */
+  long int  cv_msbp;         /* max number of steps between lsetip calls */
+  realtype  cv_dgmax_lsetup; /* gamma ratio threshold to signal for a linear
+                              * solver setup */
 
   /*------------
     Saved Values
@@ -241,11 +397,15 @@ typedef struct CVodeMemRec {
   realtype cv_h0u;             /* actual initial stepsize                     */
   realtype cv_hu;              /* last successful h value used                */
   realtype cv_saved_tq5;       /* saved value of tq[5]                        */
-  booleantype cv_jcur;         /* is Jacobian info. for lin. solver current?  */
+  booleantype cv_jcur;         /* is Jacobian info for linear solver current? */
   realtype cv_tolsf;           /* tolerance scale factor                      */
-  int cv_qmax_alloc;           /* value of qmax used when allocating memory   */
+  int cv_qmax_alloc;           /* value of qmax used when allocating mem      */
   int cv_indx_acor;            /* index of the zn vector with saved acor      */
 
+  /*--------------------------------------------------------------------
+    Flags turned ON by CVodeInit and read by CVodeReInit
+    --------------------------------------------------------------------*/
+
   booleantype cv_VabstolMallocDone;
   booleantype cv_MallocDone;
   booleantype cv_constraintsMallocDone;
@@ -258,6 +418,12 @@ typedef struct CVodeMemRec {
   void *cv_eh_data;           /* data pointer passed to ehfun                 */
   FILE *cv_errfp;             /* CVODE error messages are sent to errfp       */
 
+  /*-------------------------------------------
+    User access function
+    -------------------------------------------*/
+  CVMonitorFn cv_monitorfun;     /* func called with CVODE mem and user data  */
+  long int cv_monitor_interval;  /* step interval to call cv_monitorfun       */
+
   /*-------------------------
     Stability Limit Detection
     -------------------------*/
@@ -282,13 +448,22 @@ typedef struct CVodeMemRec {
   realtype *cv_ghi;        /* saved array of g values at t = thi              */
   realtype *cv_grout;      /* array of g values at t = trout                  */
   realtype cv_toutc;       /* copy of tout (if NORMAL mode)                   */
-  realtype cv_ttol;        /* tolerance on root location                      */
+  realtype cv_ttol;        /* tolerance on root location trout                */
   int cv_taskc;            /* copy of parameter itask                         */
   int cv_irfnd;            /* flag showing whether last step had a root       */
   long int cv_nge;         /* counter for g evaluations                       */
   booleantype *cv_gactive; /* array with active/inactive event functions      */
   int cv_mxgnull;          /* number of warning messages about possible g==0  */
 
+  /*---------------
+    Projection Data
+    ---------------*/
+
+  CVodeProjMem proj_mem;      /* projection memory structure               */
+  booleantype  proj_enabled;  /* flag indicating if projection is enabled  */
+  booleantype  proj_applied;  /* flag indicating if projection was applied */
+  realtype     proj_p[L_MAX]; /* coefficients of p(x) (degree q poly)      */
+
   /*-----------------------
     Fused Vector Operations
     -----------------------*/
@@ -296,6 +471,8 @@ typedef struct CVodeMemRec {
   realtype cv_cvals[L_MAX]; /* array of scalars */
   N_Vector cv_Xvecs[L_MAX]; /* array of vectors */
 
+  booleantype cv_usefused;  /* flag indicating if CVODE specific fused kernels should be used */
+
 } *CVodeMem;
 
 /*
@@ -312,24 +489,24 @@ typedef struct CVodeMemRec {
  *
  * CV_NO_FAILURES : Either this is the first cv_setup call for this
  *                  step, or the local error test failed on the
- *                  previous attempt at this step (but the Newton
- *                  iteration converged).
+ *                  previous attempt at this step (but the nonlinear
+ *                  solver iteration converged).
  *
  * CV_FAIL_BAD_J  : This value is passed to cv_lsetup if
  *
- *                  (a) The previous Newton corrector iteration
+ *                  (a) The previous nonlinear solver corrector iteration
  *                      did not converge and the linear solver's
  *                      setup routine indicated that its Jacobian-
  *                      related data is not current
  *                                   or
- *                  (b) During the previous Newton corrector
+ *                  (b) During the previous nonlinear solver corrector
  *                      iteration, the linear solver's solve routine
  *                      failed in a recoverable manner and the
  *                      linear solver's setup routine indicated that
  *                      its Jacobian-related data is not current.
  *
  * CV_FAIL_OTHER  : During the current internal step try, the
- *                  previous Newton iteration failed to converge
+ *                  previous nonlinear solver iteration failed to converge
  *                  even though the linear solver was using current
  *                  Jacobian-related data.
  * -----------------------------------------------------------------
@@ -432,7 +609,7 @@ typedef struct CVodeMemRec {
 
 /*
  * =================================================================
- *   C V O D E    I N T E R N A L   F U N C T I O N S
+ *    I N T E R N A L   F U N C T I O N S
  * =================================================================
  */
 
@@ -442,22 +619,36 @@ int cvEwtSet(N_Vector ycur, N_Vector weight, void *data);
 
 /* High level error handler */
 
-void cvProcessError(CVodeMem cv_mem,
-		    int error_code, const char *module, const char *fname,
-		    const char *msgfmt, ...);
+void cvProcessError(CVodeMem cv_mem, int error_code, const char *module,
+                    const char *fname, const char *msgfmt, ...);
 
 /* Prototype of internal ErrHandler function */
 
 void cvErrHandler(int error_code, const char *module, const char *function,
-		  char *msg, void *data);
+                  char *msg, void *data);
 
-/* Nonlinear solver initializtion function */
+/* Nonlinear solver initialization */
 
 int cvNlsInit(CVodeMem cv_mem);
 
+/* Projection functions */
+
+int cvDoProjection(CVodeMem cv_mem, int *nflagPtr, realtype saved_t,
+                   int *npfPtr);
+int cvProjInit(CVodeProjMem proj_mem);
+int cvProjFree(CVodeProjMem *proj_mem);
+
+/* Restore tn and undo prediction to reattempt a step */
+
+void cvRestore(CVodeMem cv_mem, realtype saved_t);
+
+/* Reset h and rescale history array to prepare for a step */
+
+void cvRescale(CVodeMem cv_mem);
+
 /*
  * =================================================================
- *   C V O D E    E R R O R    M E S S A G E S
+ *    E R R O R    M E S S A G E S
  * =================================================================
  */
 
@@ -493,6 +684,7 @@ int cvNlsInit(CVodeMem cv_mem);
 #define MSGCV_CVMEM_FAIL "Allocation of cvode_mem failed."
 #define MSGCV_MEM_FAIL "A memory request failed."
 #define MSGCV_BAD_LMM  "Illegal value for lmm. The legal values are CV_ADAMS and CV_BDF."
+#define MSGCV_NULL_SUNCTX  "sunctx = NULL illegal."
 #define MSGCV_NO_MALLOC "Attempt to call before CVodeInit."
 #define MSGCV_NEG_MAXORD "maxord <= 0 illegal."
 #define MSGCV_BAD_MAXORD  "Illegal attempt to increase maximum method order."
@@ -551,6 +743,15 @@ int cvNlsInit(CVodeMem cv_mem);
 #define MSGCV_NLS_INPUT_NULL "At " MSG_TIME ", the nonlinear solver was passed a NULL input."
 #define MSGCV_NLS_FAIL "At " MSG_TIME ", the nonlinear solver failed in an unrecoverable manner."
 
+/* CVode Projection Error Messages */
+
+#define MSG_CV_MEM_NULL  "cvode_mem = NULL illegal."
+#define MSG_CV_MEM_FAIL  "A memory request failed."
+
+#define MSG_CV_PROJ_MEM_NULL       "proj_mem = NULL illegal."
+#define MSG_CV_PROJFUNC_FAIL       "At " MSG_TIME " the projection function failed with an unrecoverable error."
+#define MSG_CV_REPTD_PROJFUNC_ERR  "At " MSG_TIME " the projection function had repeated recoverable errors."
+
 #ifdef __cplusplus
 }
 #endif
